import copy
from sympy import symbols, solve, Float
from func_timeout import func_set_timeout, FunctionTimedOut
from formalgeo.parse import get_equation_from_tree
from formalgeo.tools import rough_equal
import warnings


class EquationKiller:
    solve_eqs = True  # whether to solve the equation in the intermediate process
    sym_simplify = True  # whether to apply symbol substitution simplification
    accurate_mode = False  # whether to use accurate mode
    solve_rank_deficient_eqs = False  # whether to solve rank deficient equations
    use_cache = False  # whether to use cache to store solved target equations
    cache_eqs = None  # <dict>, {tuple(str(eqs),): [(sym_str, value)]}
    cache_target = None  # <dict>, {tuple(str(eqs),): value}

    @staticmethod
    def get_minimum_target_equations(target_expr, eqs):
        """
        Return minimum target equations. Called by function <EquationKiller.solve_target>.
        :param target_expr: Target Expression.
        :param eqs: Existing Equations.
        :return target_sym: Target symbols.
        :return mini_eqs: minimum equations rank by solving difficulty.
        :return n_m: number of equations and syms.
        """
        target_sym = symbols("t_s")
        eqs = [target_sym - target_expr] + eqs

        sym_to_eqs = {}  # dict, sym: [equation]
        for eq in eqs:
            for sym in eq.free_symbols:
                if sym in sym_to_eqs:
                    sym_to_eqs[sym].append(eq)
                else:
                    sym_to_eqs[sym] = [eq]

        mini_eqs = [eqs[0]]  # mini equations
        mini_syms = eqs[0].free_symbols  # sym of mini equations
        n_m = [(len(mini_eqs), len(mini_syms))]  # number of equations and variable

        related_eqs = []  # related eqs waiting to add
        for sym in mini_syms:
            for r_eq in sym_to_eqs[sym]:
                related_eqs.append(r_eq)
        related_eqs = list(set(related_eqs) - set(mini_eqs))

        while len(related_eqs) > 0:
            added_eq_id = 0
            added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
            added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)
            for i in range(1, len(related_eqs)):
                if len(related_eqs[i].free_symbols - mini_syms) < added_eq_n1:
                    added_eq_id = i
                    added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
                    added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)
                elif len(related_eqs[i].free_symbols - mini_syms) == added_eq_n1:
                    if len(related_eqs[i].free_symbols) > added_eq_n2:
                        added_eq_id = i
                        added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
                        added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)

            added_eq = related_eqs[added_eq_id]
            mini_eqs.append(added_eq)
            mini_syms |= added_eq.free_symbols
            n_m.append((len(mini_eqs), len(mini_syms)))

            for sym in added_eq.free_symbols:
                for r_eq in sym_to_eqs[sym]:
                    related_eqs.append(r_eq)
            related_eqs = list(set(related_eqs) - set(mini_eqs))

        return target_sym, mini_eqs, n_m

    @staticmethod
    def get_minimum_group_equations(eqs):
        """
        Return minimum group equations. Called by function <EquationKiller.solve_equations>.
        :param eqs: Equations.
        :return mini_eqs_list: minimum equations lists rank by solving difficulty.
        :return n_m: number of equations and syms.
        """

        sym_to_eqs = {}  # dict, sym: [equation]
        for eq in eqs:
            for sym in eq.free_symbols:
                if sym in sym_to_eqs:
                    sym_to_eqs[sym].append(eq)
                else:
                    sym_to_eqs[sym] = [eq]

        mini_eqs_lists = []  # mini equations
        n_m = []  # number of equations and variable

        added_eqs = set()
        for eq in eqs:
            if eq in added_eqs:
                continue
            added_eqs.add(eq)

            mini_eqs = [eq]  # mini equations
            mini_syms = eq.free_symbols  # sym of mini equations

            related_eqs = []  # related eqs waiting to add
            for sym in mini_syms:
                for r_eq in sym_to_eqs[sym]:
                    related_eqs.append(r_eq)
            related_eqs = list(set(related_eqs) - set(mini_eqs))

            if len(related_eqs) == 0:
                mini_eqs_lists.append(mini_eqs)
                n_m.append((len(mini_eqs), len(mini_syms)))  # add mini equations
                continue

            while True:
                added_eq_id = 0
                added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
                added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)
                for i in range(1, len(related_eqs)):
                    if len(related_eqs[i].free_symbols - mini_syms) < added_eq_n1:
                        added_eq_id = i
                        added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
                        added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)
                    elif len(related_eqs[i].free_symbols - mini_syms) == added_eq_n1:
                        if len(related_eqs[i].free_symbols) > added_eq_n2:
                            added_eq_id = i
                            added_eq_n1 = len(related_eqs[added_eq_id].free_symbols - mini_syms)
                            added_eq_n2 = len(related_eqs[added_eq_id].free_symbols)

                added_eq = related_eqs[added_eq_id]
                mini_eqs.append(added_eq)
                mini_syms |= added_eq.free_symbols
                added_eqs.add(added_eq)

                for sym in added_eq.free_symbols:
                    for r_eq in sym_to_eqs[sym]:
                        related_eqs.append(r_eq)
                related_eqs = list(set(related_eqs) - set(mini_eqs))

                if len(related_eqs) == 0:
                    mini_eqs_lists.append(mini_eqs)
                    n_m.append((len(mini_eqs), len(mini_syms)))  # add mini equations
                    break

        return mini_eqs_lists, n_m

    @staticmethod
    def get_minimum_syms(target_eqs, eqs):
        """
        Return minimum equation's syms. Called by function <Searcher.get_theorem_selection>.
        :param target_eqs: <list>, target Equations.
        :param eqs: <list>, existing Equations.
        :return syms: <set>, set of minimum equation's syms.
        """
        sym_to_eqs = {}  # dict, sym: [equation]
        for eq in target_eqs + eqs:
            for sym in eq.free_symbols:
                if sym in sym_to_eqs:
                    sym_to_eqs[sym].append(eq)
                else:
                    sym_to_eqs[sym] = [eq]

        mini_eqs = set(target_eqs)
        mini_syms = set()
        for eq in mini_eqs:
            mini_syms |= eq.free_symbols

        while True:
            new_sym = set()
            for sym in mini_syms:
                for eq in sym_to_eqs[sym]:
                    mini_eqs.add(eq)
                    new_sym |= eq.free_symbols
            new_sym = new_sym - mini_syms
            if len(new_sym) == 0:
                break
            mini_syms |= new_sym

        return mini_syms

    @staticmethod
    @func_set_timeout(2)
    def simplification_value_replace(problem):
        """
        Simplify equations by replacing sym with known value.
        :param problem: Instance of class <Problem>.
        """
        update = True
        while update:
            update = False
            remove_lists = set()  # equation to be deleted
            add_lists = []  # equation to be added

            for eq in problem.condition.simplified_equation:  # solve eq that only one sym unsolved
                if len(eq.free_symbols) != 1:
                    continue

                target_sym = list(eq.free_symbols)[0]
                try:
                    result = EquationKiller.solve(eq)  # solve equations
                except FunctionTimedOut:
                    msg = "Timeout when solve equation: {}".format(eq)
                    warnings.warn(msg)
                else:
                    if target_sym in result:
                        problem.set_value_of_sym(target_sym, result[target_sym],
                                                 tuple(problem.condition.simplified_equation[eq]))
                        remove_lists |= {eq}

            for eq in problem.condition.simplified_equation:  # value replace
                if eq in remove_lists:
                    continue
                raw_eq = eq
                simplified = False
                added_premise = []
                for sym in eq.free_symbols:
                    if problem.condition.value_of_sym[sym] is None:
                        continue
                    simplified = True  # replace sym with value when the value known
                    eq = eq.subs(sym, problem.condition.value_of_sym[sym])
                    added_premise.append(problem.condition.get_id_by_predicate_and_item(
                        "Equation", sym - problem.condition.value_of_sym[sym]))
                    remove_lists |= {raw_eq}

                if not simplified:
                    continue

                if len(eq.free_symbols) == 0:  # no need to add new simplified equation when it's all sym known
                    continue
                else:  # add new simplified equation
                    update = True
                    premise = problem.condition.simplified_equation[raw_eq] + added_premise
                    add_lists.append((eq, premise))

            for remove_eq in remove_lists:  # remove useless equation
                problem.condition.simplified_equation.pop(remove_eq)
            for add_eq, premise in add_lists:  # remove useless equation
                problem.condition.simplified_equation[add_eq] = premise

    @staticmethod
    @func_set_timeout(2)
    def simplification_sym_replace(equations, target_sym):
        """ High level simplify based on symbol replacement."""
        update = True
        while update:
            update = False
            for i in range(len(equations)):
                eq = equations[i]

                if target_sym in eq.free_symbols or \
                        len(eq.free_symbols) != 2 or \
                        len(eq.atoms()) > 5:  # too many atoms, no need to replace
                    continue

                try:
                    result = EquationKiller.solve(eq, keep_sym=True)  # solve sym
                except FunctionTimedOut:
                    msg = "Timeout when solve equations: {}".format(equations)
                    warnings.warn(msg)
                    continue

                if len(result) == 0:  # no solved result
                    continue

                sym = list(result.keys())[0]
                has_float = False
                for atom in result[sym].atoms():
                    if isinstance(atom, Float):
                        has_float = True
                        break
                if has_float:  # float has precision error
                    continue

                for j in range(len(equations)):  # replace sym with solved sym_expr
                    if sym in equations[j].free_symbols:
                        equations[j] = equations[j].subs(sym, result[sym])
                        update = True

        for i in range(len(equations))[::-1]:  # remove 0
            if len(equations[i].free_symbols) == 0:
                equations.pop(i)

    @staticmethod
    @func_set_timeout(2)
    def solve(equations, target_sym=None, keep_sym=False):
        try:
            if target_sym is not None:
                solved = solve(equations, target_sym, dict=True)
            else:
                solved = solve(equations, dict=True)

            if len(solved) == 0:  # no result solved
                return {}
        except Exception as e:  # exception
            msg = "Exception <{}> occur when solve {}".format(e, equations)
            warnings.warn(msg)
            return {}
        else:  # has result
            if keep_sym:  # keep sym result
                if isinstance(solved, list):
                    return solved[0]
                return solved

            if isinstance(solved, list):
                update = True
                while update and len(solved) > 1:  # choose min but > 0, when has multi result
                    update = False
                    for sym in solved[0]:
                        if sym not in solved[1]:
                            solved.pop(1)
                            update = True
                            break
                        if len(solved[0][sym].free_symbols) != 0 and len(solved[1][sym].free_symbols) == 0:
                            solved.pop(0)
                            update = True
                            break
                        if len(solved[0][sym].free_symbols) == 0 and len(solved[1][sym].free_symbols) != 0:
                            solved.pop(1)
                            update = True
                            break
                        if len(solved[0][sym].free_symbols) != 0:
                            continue
                        if solved[0][sym] == solved[1][sym]:
                            continue
                        if float(solved[0][sym]) < 0 < float(solved[1][sym]):
                            solved.pop(0)
                            update = True
                            break
                        if float(solved[0][sym]) > 0 > float(solved[1][sym]):
                            solved.pop(1)
                            update = True
                            break

                    if update:
                        continue

                    for sym in solved[0]:
                        if len(solved[0][sym].free_symbols) != 0:
                            continue
                        if solved[0][sym] == solved[1][sym]:
                            continue
                        if abs(float(solved[0][sym])) > abs(float(solved[1][sym])):
                            solved.pop(0)
                            update = True
                        else:
                            solved.pop(1)
                            update = True
                        break

                solved = solved[0]

            real_results = {}  # real_number
            for sym in solved:  # filter out real number solution
                if len(solved[sym].free_symbols) == 0:
                    # real_results[sym] = number_round(solved[sym])
                    real_results[sym] = solved[sym]
            return real_results

    @staticmethod
    def solve_equations(problem):
        """
        Solve equations in problem.condition.equations.
        :param problem: Instance of class <Problem>.
        """
        if not EquationKiller.solve_eqs or problem.condition.eq_solved:
            return

        try:
            EquationKiller.simplification_value_replace(problem)  # simplify equations before solving
        except FunctionTimedOut:
            msg = "Timeout when simplify equations by value replace."
            warnings.warn(msg)

        mini_eqs_lists, n_m = EquationKiller.get_minimum_group_equations(  # get mini equations
            list(problem.condition.simplified_equation)
        )

        for i in range(len(mini_eqs_lists)):
            if not EquationKiller.solve_rank_deficient_eqs and n_m[i][0] < n_m[i][1]:
                continue

            eqs_for_cache = None
            if EquationKiller.use_cache:
                eqs_for_cache = []
                premise = []
                str_to_sym = {}
                for eq in mini_eqs_lists[i]:
                    eqs_for_cache.append(str(eq))
                    premise += problem.condition.simplified_equation[eq]
                    for sym in eq.free_symbols:
                        str_to_sym[str(sym)] = sym
                eqs_for_cache = tuple(sorted(eqs_for_cache))

                if eqs_for_cache in EquationKiller.cache_eqs:
                    for sym_str, value in EquationKiller.cache_eqs[eqs_for_cache]:
                        problem.set_value_of_sym(str_to_sym[sym_str], value, premise)
                    continue
                EquationKiller.cache_eqs[eqs_for_cache] = []

            solved = False
            solved_results = None
            mini_eqs = None

            try:
                results = EquationKiller.solve(mini_eqs_lists[i])  # solve equations
            except FunctionTimedOut:
                msg = "Timeout when solve equations: {}".format(mini_eqs_lists[i])
                warnings.warn(msg)
            else:
                for sym in results:
                    if problem.condition.value_of_sym[sym] is None:
                        solved = True
                        solved_results = results
                        mini_eqs = mini_eqs_lists[i]
                        break

            if not solved:
                continue

            if EquationKiller.accurate_mode:
                for sym in solved_results:
                    if problem.condition.value_of_sym[sym] is not None:
                        continue
                    sym_mini_eqs = copy.copy(mini_eqs)
                    for removed_eq in copy.copy(sym_mini_eqs):
                        try_eqs = copy.copy(sym_mini_eqs)
                        try_eqs.remove(removed_eq)
                        try:
                            results = EquationKiller.solve(try_eqs, sym)  # solve equations
                        except FunctionTimedOut:
                            msg = "Timeout when solve equations: {}".format(try_eqs)
                            warnings.warn(msg)
                        else:
                            if sym in results:
                                sym_mini_eqs.remove(removed_eq)

                    premise = []
                    for eq in sym_mini_eqs:
                        premise += problem.condition.simplified_equation[eq]
                    problem.set_value_of_sym(sym, solved_results[sym], premise)

                    if EquationKiller.use_cache:
                        EquationKiller.cache_eqs[eqs_for_cache].append((str(sym), solved_results[sym]))

            else:
                premise = []
                for eq in mini_eqs:
                    premise += problem.condition.simplified_equation[eq]

                for sym in solved_results:
                    problem.set_value_of_sym(sym, solved_results[sym], premise)

                if EquationKiller.use_cache:
                    for sym in solved_results:
                        EquationKiller.cache_eqs[eqs_for_cache].append((str(sym), solved_results[sym]))

    @staticmethod
    def solve_target(target_expr, problem):
        """
        Solve target_expr in the constraint of problem's equation.
        :param problem: Instance of class <Problem>.
        :param target_expr: symbol expression.
        """
        if target_expr is None:
            return None, None

        if target_expr in problem.condition.get_items_by_predicate("Equation"):  # no need to solve
            return 0, [problem.condition.get_id_by_predicate_and_item("Equation", target_expr)]
        if -target_expr in problem.condition.get_items_by_predicate("Equation"):
            return 0, [problem.condition.get_id_by_predicate_and_item("Equation", -target_expr)]

        try:
            EquationKiller.simplification_value_replace(problem)  # simplify equations before solving
        except FunctionTimedOut:
            msg = "Timeout when simplify equations by value replace."
            warnings.warn(msg)

        premise = []
        for sym in target_expr.free_symbols:  # solve only using value replacement
            if problem.condition.value_of_sym[sym] is not None:
                target_expr = target_expr.subs(sym, problem.condition.value_of_sym[sym])
                premise.append(problem.condition.get_id_by_predicate_and_item(
                    "Equation", sym - problem.condition.value_of_sym[sym]))
        if len(target_expr.free_symbols) == 0:
            return target_expr, premise

        target_sym, mini_eqs, n_m = EquationKiller.get_minimum_target_equations(  # get mini equations
            target_expr,
            list(problem.condition.simplified_equation)
        )

        if len(mini_eqs) == 0:  # no mini equations, can't solve
            return None, []

        eqs_for_cache = None
        if EquationKiller.use_cache:
            eqs_for_cache = [str(mini_eqs[0])]
            for eq in mini_eqs[1:]:
                eqs_for_cache.append(str(eq))
                premise += problem.condition.simplified_equation[eq]
            eqs_for_cache = tuple(sorted(eqs_for_cache))

            if eqs_for_cache in EquationKiller.cache_target:
                value = EquationKiller.cache_target[eqs_for_cache]
                if value is None:
                    return None, []
                return value, premise
            EquationKiller.cache_target[eqs_for_cache] = None

        head = 0  # can't solve
        tail = len(mini_eqs)  # can solve
        solved_mini_eqs = None
        solved_target_value = None
        while tail - head > 1:
            solved = False
            p = int((head + tail) / 2)
            try_mini_eqs = copy.copy(mini_eqs[0:p + 1])

            if EquationKiller.sym_simplify:
                try:
                    EquationKiller.simplification_sym_replace(try_mini_eqs, target_sym)
                except FunctionTimedOut:
                    msg = "Timeout when simplify equations by sym replace."
                    warnings.warn(msg)

            try:
                results = EquationKiller.solve(try_mini_eqs)  # solve equations
            except FunctionTimedOut:
                msg = "Timeout when solve equations: {}".format(try_mini_eqs)
                warnings.warn(msg)
            else:
                if target_sym in results:
                    solved = True
                    solved_mini_eqs = copy.copy(mini_eqs[0:p + 1])
                    solved_target_value = results[target_sym]

            if solved:
                tail = p
                if not EquationKiller.accurate_mode:
                    break
            else:
                head = p

        if solved_target_value is None:  # no solved result
            return None, []

        if EquationKiller.accurate_mode:
            for removed_eq in copy.copy(solved_mini_eqs[1:]):
                try_mini_eqs = copy.copy(solved_mini_eqs)
                try_mini_eqs.remove(removed_eq)

                if EquationKiller.sym_simplify:
                    try:
                        EquationKiller.simplification_sym_replace(try_mini_eqs, target_sym)
                    except FunctionTimedOut:
                        msg = "Timeout when simplify equations by sym replace."
                        warnings.warn(msg)

                try:
                    results = EquationKiller.solve(try_mini_eqs, target_sym)  # solve equations
                except FunctionTimedOut:
                    msg = "Timeout when solve equations: {}".format(try_mini_eqs)
                    warnings.warn(msg)
                else:
                    if target_sym in results:
                        solved_mini_eqs.remove(removed_eq)

        for eq in solved_mini_eqs[1:]:
            premise += problem.condition.simplified_equation[eq]

        eq = target_expr - solved_target_value
        value_added = False
        if len(eq.free_symbols) == 1:
            try:
                results = EquationKiller.solve(eq, list(eq.free_symbols)[0])  # solve equations
            except FunctionTimedOut:
                msg = "Timeout when solve equations: {}".format(target_expr - solved_target_value)
                warnings.warn(msg)
            else:
                for sym in results:
                    problem.set_value_of_sym(sym, results[sym], premise)
                    value_added = True
        if not value_added:
            problem.condition.add("Equation", target_expr - solved_target_value, premise, ("solve_eq", None, None))

        if EquationKiller.use_cache:
            EquationKiller.cache_target[eqs_for_cache] = solved_target_value
        return solved_target_value, premise


class GeometryPredicateLogicExecutor:

    @staticmethod
    def run(gpl, problem, letters=None):
        """
        Run reason step by step.
        :param gpl: <dict>, (products, logic_constraints, algebra_constraints, conclusions), geometric predicate logic.
        :param problem: instance of class <Problem>.
        :param letters: preset letters for para selection.
        :return results: <list> of <tuple>, [(letters, premises, conclusions)].
        """
        r = GeometryPredicateLogicExecutor.run_logic(gpl, problem, letters)
        r = GeometryPredicateLogicExecutor.run_algebra(r, gpl, problem)
        return GeometryPredicateLogicExecutor.make_conclusion(r, gpl, problem)

    @staticmethod
    def run_logic(gpl, problem, letters=None):
        """
        Run 'products', 'logic_constraints' of GPL.
        :param gpl: <dict>, (products, logic_constraints, algebra_constraints, conclusions), geometric predicate logic.
        :param problem: instance of class <Problem>.
        :param letters: preset letters for para selection.
        :return r: triplet, (r_ids, r_items, r_vars).
        """
        products = gpl["products"]
        logic_constraints = gpl["logic_constraints"]

        r_ids, r_items = problem.condition.get_ids_and_items_by_predicate_and_variable(products[0][0], products[0][1])
        r_vars = products[0][1]
        for i in range(len(r_items)):  # delete duplicated vars and corresponding item
            r_items[i] = list(r_items[i])
        r_vars = list(r_vars)
        deleted_vars_index = []  # deleted vars index
        for i in range(len(r_vars)):
            if r_vars[i] in r_vars[0:i]:
                deleted_vars_index.append(i)
        for index in deleted_vars_index[::-1]:  # delete
            r_vars.pop(index)
            for i in range(len(r_items)):
                r_items[i].pop(index)

        for i in range(1, len(products)):
            r_ids, r_items, r_vars = GeometryPredicateLogicExecutor.product(
                (r_ids, r_items, r_vars), products[i], problem)

        if letters is not None:  # select result according to letters
            for i in range(len(r_ids))[::-1]:
                selected = True
                for v in letters:
                    if r_items[i][r_vars.index(v)] != letters[v]:
                        selected = False
                        break
                if not selected:
                    r_items.pop(i)
                    r_ids.pop(i)

        for i in range(len(logic_constraints)):
            r_ids, r_items, r_vars = GeometryPredicateLogicExecutor.constraint_logic(
                (r_ids, r_items, r_vars), logic_constraints[i], problem)

        if letters is not None:  # select result according to letters
            for i in range(len(r_ids))[::-1]:
                selected = True
                for v in letters:
                    if r_items[i][r_vars.index(v)] != letters[v]:
                        selected = False
                        break
                if not selected:
                    r_items.pop(i)
                    r_ids.pop(i)

        return r_ids, r_items, r_vars

    @staticmethod
    def run_algebra(r, gpl, problem):
        """
        Run 'algebra_constraints' of GPL.
        :param r: triplet, (r_ids, r_items, r_vars).
        :param gpl: <dict>, (products, logic_constraints, algebra_constraints, conclusions), geometric predicate logic.
        :param problem: instance of class <Problem>.
        :return results: <list> of <tuple>, [(letters, premises, conclusions)].
        """
        algebra_constraints = gpl["algebra_constraints"]
        r_ids, r_items, r_vars = r
        if len(r_ids) == 0:
            return [], [], r_vars

        for i in range(len(algebra_constraints)):
            r_ids, r_items, r_vars = GeometryPredicateLogicExecutor.constraint_algebra(
                (r_ids, r_items, r_vars), algebra_constraints[i], problem)

        return r_ids, r_items, r_vars

    @staticmethod
    def make_conclusion(r, gpl, problem):
        """
        Make conclusion according given reasoned points sets 'r' and GDL 'conclusions'.
        :param r: triplet, (r_ids, r_items, r_vars).
        :param gpl: <dict>, (products, logic_constraints, algebra_constraints, conclusions), geometric predicate logic.
        :param problem: instance of class <Problem>.
        :return results: <list> of <tuple>, [(letters, premises, conclusions)].
        """
        if len(r[0]) == 0:
            return []
        conclusions = gpl["conclusions"]
        results = []
        r_ids, r_items, r_vars = r
        for i in range(len(r_ids)):
            letters = {}
            for j in range(len(r_vars)):
                letters[r_vars[j]] = r_items[i][j]
            conclusion = []

            for predicate, item in conclusions:
                if predicate == "Equal":  # algebra conclusion
                    eq = get_equation_from_tree(problem, item, True, letters)
                    conclusion.append(("Equation", eq))
                else:  # logic conclusion
                    item = tuple(letters[i] for i in item)
                    conclusion.append((predicate, item))
            results.append((letters, r_ids[i], conclusion))

        return results

    @staticmethod
    def product(r1, r2_logic, problem):
        """
        Constrained Cartesian product.
        :param r1: triplet, (r1_ids, r1_items, r1_vars).
        :param r2_logic: geo predicate logic, such as ['Collinear', ['a', 'b', 'c']].
        :param problem: instance of class <Problem>.
        :return r: triplet, (r_ids, r_items, r_vars), reasoning result.
        >> product(([(1,), (2,)], [('A', 'B'), ('C', 'D')], ['a', 'b']),
                   ['Line', ['b', 'c']],
                   problem)
        ([(1, 3), (2, 4)], [('A', 'B', 'C'), ('C', 'D', 'E')], ['a', 'b', 'c'])
        """
        r1_ids, r1_items, r1_vars = r1
        if len(r1_ids) == 0:
            return [], [], r1_vars
        r2_ids, r2_items = problem.condition.get_ids_and_items_by_predicate_and_variable(r2_logic[0], r2_logic[1])
        r2_vars = r2_logic[1]

        inter = list(set(r1_vars) & set(r2_vars))  # intersection
        for i in range(len(inter)):
            inter[i] = (r1_vars.index(inter[i]), r2_vars.index(inter[i]))  # change to index

        difference = list(set(r2_vars) - set(r1_vars))  # difference
        for i in range(len(difference)):
            difference[i] = r2_vars.index(difference[i])  # change to index

        r_ids = []  # result
        r_items = []
        r_vars = list(r1_vars)
        for dif in difference:  # add r2 vars
            r_vars.append(r2_vars[dif])
        r_vars = tuple(r_vars)

        for i in range(len(r1_items)):
            r1_data = r1_items[i]
            for j in range(len(r2_items)):
                r2_data = r2_items[j]
                passed = True
                for r1_i, r2_i in inter:
                    if r1_data[r1_i] != r2_data[r2_i]:  # the corresponding points are inconsistent.
                        passed = False
                        break
                if passed:
                    item = list(r1_data)
                    for dif in difference:
                        item.append(r2_data[dif])
                    r_items.append(tuple(item))
                    r_ids.append(tuple(set(list(r1_ids[i]) + list(r2_ids[j]))))
        return r_ids, r_items, r_vars

    @staticmethod
    def constraint_logic(r1, r2_logic, problem):
        """
        Logic constraint.
        :param r1: triplet, (r1_ids, r1_items, r1_vars).
        :param r2_logic: geo predicate logic, such as ['Collinear', ['a', 'b', 'c']].
        :param problem: instance of class <Problem>.
        :return r: triplet, (r_ids, r_items, r_vars), reasoning result.
        >> problem.conditions['Line'].get_item_by_id  # supposed
        {3: ('B', 'C')}
        >> constraint_logic(([(1,), (2,)], [('A', 'B', 'C'), ('C', 'D', 'E')], ['a', 'b', 'c']),
                            ['Line', ['b', 'c']],
                            problem)
        ([(1, 3)], [('A', 'B', 'C')], ['a', 'b', 'c'])
        >> constraint_logic(([(1,), (2,)], [('A', 'B', 'C'), ('C', 'D', 'E')], ['a', 'b', 'c']),
                            ['~Line', ['b', 'c']],
                            problem)
        ([(2,)], [('C', 'D', 'E')], ['a', 'b', 'c'])
        """
        r1_ids, r1_items, r1_vars = r1
        if len(r1_ids) == 0:
            return [], [], r1_vars
        oppose = False  # indicate '&' or '&~'
        if "~" in r2_logic[0]:
            r2_logic = list(r2_logic)
            r2_logic[0] = r2_logic[0].replace("~", "")
            r2_logic = tuple(r2_logic)
            oppose = True
        index = [r1_vars.index(v) for v in r2_logic[1]]
        r_ids = []
        r_items = []

        if not oppose:  # &
            for i in range(len(r1_items)):
                r2_item = tuple(r1_items[i][j] for j in index)
                if r2_item in problem.condition.get_items_by_predicate(r2_logic[0]):
                    r2_id = problem.condition.get_id_by_predicate_and_item(r2_logic[0], r2_item)
                    r_ids.append(tuple(set(list(r1_ids[i]) + [r2_id])))
                    r_items.append(r1_items[i])
        else:  # &~
            for i in range(len(r1_items)):
                r2_item = tuple(r1_items[i][j] for j in index)
                if r2_item not in problem.condition.get_items_by_predicate(r2_logic[0]):
                    r_ids.append(r1_ids[i])
                    r_items.append(r1_items[i])
        return r_ids, r_items, r1_vars

    @staticmethod
    def constraint_algebra(r1, r2_algebra, problem):
        """
        Algebra constraint.
        :param r1: triplet, (r1_ids, r1_items, r1_vars).
        :param r2_algebra: geo predicate logic, such as ['Equal', [['Length', ['a', 'b']], 5]].
        :param problem: instance of class <Problem>.
        :return r: triplet, (r_ids, r_items, r_vars), reasoning result.
        >> problem.conditions['Equation'].get_value_of_sym  # supposed
        {ll_ab: 1}
        >> problem.conditions['Equation'].get_item_by_id  # supposed
        {3: ll_ab - 1}
        >> constraint_algebra(([(1,), (2,)], [('A', 'B', 'C'), ('C', 'D', 'E')], ['a', 'b', 'c']),
                              ['Equal', [['Length', ['a', 'b']], 1]],
                              problem)
        ([(1, 3)], [('A', 'B', 'C')], ['a', 'b', 'c'])
        >> constraint_algebra(([(1,), (2,)], [('A', 'B', 'C'), ('C', 'D', 'E')], ['a', 'b', 'c']),
                              ['~Equal', [['Length', ['a', 'b']], 1]],
                              problem)
        ([(2,)], [('C', 'D', 'E')], ['a', 'b', 'c'])
        """
        r1_ids, r1_items, r1_vars = r1
        if len(r1_ids) == 0:
            return [], [], r1_vars
        oppose = False  # indicate '&' or '&~'
        if "~" in r2_algebra[0]:
            r2_algebra = list(r2_algebra)
            r2_algebra[0] = r2_algebra[0].replace("~", "")
            r2_algebra = tuple(r2_algebra)
            oppose = True
        r_ids = []
        r_items = []

        if not oppose:  # &
            for i in range(len(r1_items)):
                letters = {}
                for j in range(len(r1_vars)):
                    letters[r1_vars[j]] = r1_items[i][j]
                eq = get_equation_from_tree(problem, r2_algebra[1], True, letters)
                try:
                    result, premise = EquationKiller.solve_target(eq, problem)
                except FunctionTimedOut:
                    msg = "Timeout when solve target: {}".format(str(eq))
                    warnings.warn(msg)
                else:
                    if result is not None and rough_equal(result, 0):  # meet constraints
                        r_id = tuple(set(premise + list(r1_ids[i])))
                        r_ids.append(r_id)
                        r_items.append(r1_items[i])

        else:  # &~
            for i in range(len(r1_items)):
                letters = {}
                for j in range(len(r1_vars)):
                    letters[r1_vars[j]] = r1_items[i][j]
                eq = get_equation_from_tree(problem, r2_algebra[1], True, letters)
                try:
                    result, premise = EquationKiller.solve_target(eq, problem)
                except FunctionTimedOut:
                    msg = "Timeout when solve target: {}".format(str(eq))
                    warnings.warn(msg)
                else:
                    if result is None or not rough_equal(result, 0):  # meet constraints
                        r_id = tuple(set(premise + list(r1_ids[i])))
                        r_ids.append(r_id)
                        r_items.append(r1_items[i])

        return r_ids, r_items, r1_vars



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""
'solver' invokes other modules to enable interactive problem-solving and automated problem-solving.
The automated problem-solving implements both forward search and backward search, allowing for the
configuration of various search strategies (breadth-first, depth-first, random, beam).
"""

__all__ = [
    "Interactor", "ForwardSearcher", "BackwardSearcher"
]

from formalgeo.solver.interactive import Interactor
from formalgeo.solver.forward_search import ForwardSearcher
from formalgeo.solver.backward_search import BackwardSearcher



import time
import random
from formalgeo.problem import Problem
from formalgeo.core import GeometryPredicateLogicExecutor as GPLExecutor
from formalgeo.core import EquationKiller as EqKiller
from formalgeo.parse import parse_predicate_gdl, parse_theorem_gdl, parse_problem_cdl
from formalgeo.tools import get_used_pid_and_theorem, debug_print


def get_p2t_map_fw(t_info, parsed_theorem_GDL):
    """
    Predicate-theorem mapping hash table for Search Accelerating.
    :param t_info: <dict>, {t_name: (category_id, usage_count)}, user customization.
    :param parsed_theorem_GDL: parsed theorem GDL.
    """
    p2t_map_fw = {}
    for t_name in t_info:
        if t_info[t_name][1] == 0 or t_info[t_name][0] == 3:  # skip no used and diff t
            continue
        for t_branch in parsed_theorem_GDL[t_name]["body"]:
            theorem_unit = parsed_theorem_GDL[t_name]["body"][t_branch]
            premises = list(theorem_unit["products"])
            premises += list(theorem_unit["logic_constraints"])
            premises += list(theorem_unit["attr_in_algebra_constraints"])
            for predicate, p_vars in premises:
                if predicate[0] == "~":  # skip oppose
                    continue
                if predicate not in p2t_map_fw:
                    p2t_map_fw[predicate] = [(t_name, t_branch, p_vars)]
                elif (t_name, t_branch, p_vars) not in p2t_map_fw[predicate]:
                    p2t_map_fw[predicate].append((t_name, t_branch, p_vars))
    return p2t_map_fw


class ForwardSearcher:

    def __init__(self, predicate_GDL, theorem_GDL, strategy, max_depth, beam_size, t_info, debug=False):
        """
        Initialize Forward Searcher.
        :param predicate_GDL: predicate GDL.
        :param theorem_GDL: theorem GDL.
        :param strategy: <str>, "dfs", "bfs", "rs", "bs".
        :param max_depth: max search depth.
        :param beam_size: beam search size.
        :param t_info: <dict>, {t_name: (category_id, usage_count)}, user customization.
        :param debug: <bool>, set True when need print process information.
        """
        self.parsed_predicate_GDL = parse_predicate_gdl(predicate_GDL)
        self.parsed_theorem_GDL = parse_theorem_gdl(theorem_GDL, self.parsed_predicate_GDL)
        self.max_depth = max_depth
        self.beam_size = beam_size
        self.strategy = strategy
        self.debug = debug
        self.p2t_map = get_p2t_map_fw(t_info, self.parsed_theorem_GDL)

        self.problem = None
        self.stack = None
        self.last_step = None
        self.step_size = None
        self.node_count = None  # {depth: node_count}

        self.problem_p_paras = None  # Perimeter
        self.problem_a_paras = None  # Area

    def init_search(self, problem_CDL):
        """Initial problem by problem_CDL and build root Node."""
        EqKiller.use_cache = True  # use cache to speed up solving
        EqKiller.cache_eqs = {}
        EqKiller.cache_target = {}

        timing = time.time()  # timing

        self.problem = Problem()  # init problem
        self.problem.load_problem_by_fl(
            self.parsed_predicate_GDL, self.parsed_theorem_GDL, parse_problem_cdl(problem_CDL))
        EqKiller.solve_equations(self.problem)
        self.problem.step("init_problem", 0)

        self.stack = []
        self.last_step = 0
        self.step_size = 0
        self.node_count = {1: 1}

        self.problem_p_paras = set()  # Perimeter
        self.problem_a_paras = set()  # Area
        for sym in self.problem.condition.attr_of_sym:
            predicate, paras = self.problem.condition.attr_of_sym[sym]
            if predicate.startswith("Perimeter"):
                for para in paras:
                    self.problem_p_paras.add(para)
            elif predicate.startswith("Area"):
                for para in paras:
                    self.problem_a_paras.add(para)

        debug_print(self.debug, "(pid={}, strategy={}, timing={:.4f}s) Initialize and start forward search...".format(
            problem_CDL["problem_id"], self.strategy, time.time() - timing))

        timing = time.time()
        selections = self.get_theorem_selection()
        self.add_selections([], selections)
        debug_print(self.debug, "(timing={:.4f}s) Expand {} child node.".
                    format(time.time() - timing, len(selections)))

    def search(self):
        """
        Search problem and return search result.
        :return solved: <bool>, indicate whether problem solved or not.
        :return seqs: <list> of <str>, solved theorem sequences.
        """
        if self.strategy == "bfs":  # breadth-first search
            while len(self.stack) > 0:
                pos, selection = self.stack.pop(0)
                self.step_size += 1
                debug_print(self.debug, "\n(pos={}, node_count={}) Current node.".format(pos, self.node_count))
                timing = time.time()
                solved = self.apply_and_check_goal(selection)
                debug_print(self.debug, "(solved={}, timing={:.4f}s) Apply selection and check goal.".format(
                    solved, time.time() - timing))
                if solved is None:  # not update, close search branch
                    continue
                if solved:  # solved, return result
                    _, seqs = get_used_pid_and_theorem(self.problem)
                    return True, seqs
                else:  # continue search
                    if len(pos) == self.max_depth:
                        continue
                    timing = time.time()
                    selections = self.get_theorem_selection()
                    self.add_selections(pos, selections)
                    debug_print(self.debug, "(timing={:.4f}s) Expand {} child node.".
                                format(time.time() - timing, len(selections)))
        elif self.strategy == "dfs":  # deep-first search
            while len(self.stack) > 0:
                pos, selection = self.stack.pop()
                self.step_size += 1
                debug_print(self.debug, "\n(pos={}, node_count={}) Current node.".format(pos, self.node_count))
                timing = time.time()
                solved = self.apply_and_check_goal(selection)
                debug_print(self.debug, "(solved={}, timing={:.4f}s) Apply selection and check goal.".format(
                    solved, time.time() - timing))
                if solved is None:  # not update, close search branch
                    continue
                if solved:  # solved, return result
                    _, seqs = get_used_pid_and_theorem(self.problem)
                    return True, seqs
                else:  # continue search
                    if len(pos) == self.max_depth:
                        continue
                    timing = time.time()
                    selections = self.get_theorem_selection()
                    self.add_selections(pos, selections)
                    debug_print(self.debug, "(timing={:.4f}s) Expand {} child node.".
                                format(time.time() - timing, len(selections)))
        elif self.strategy == "rs":  # random search
            while len(self.stack) > 0:
                pos, selection = self.stack.pop(random.randint(0, len(self.stack) - 1))
                self.step_size += 1
                debug_print(self.debug, "\n(pos={}, node_count={}) Current node.".format(pos, self.node_count))
                timing = time.time()
                solved = self.apply_and_check_goal(selection)
                debug_print(self.debug, "(solved={}, timing={:.4f}s) Apply selection and check goal.".format(
                    solved, time.time() - timing))
                if solved is None:  # not update, close search branch
                    continue
                if solved:  # solved, return result
                    _, seqs = get_used_pid_and_theorem(self.problem)
                    return True, seqs
                else:  # continue search
                    if len(pos) == self.max_depth:
                        continue
                    timing = time.time()
                    selections = self.get_theorem_selection()
                    self.add_selections(pos, selections)
                    debug_print(self.debug, "(timing={:.4f}s) Expand {} child node.".
                                format(time.time() - timing, len(selections)))
        else:  # beam search
            while len(self.stack) > 0:
                beam_count = len(self.stack)
                if len(self.stack) > self.beam_size:  # select branch with beam size
                    stack = []
                    for i in random.sample(range(len(self.stack)), self.beam_size):
                        stack.append(self.stack[i])
                    self.stack = stack
                    beam_count = self.beam_size

                for i in range(beam_count):
                    pos, selection = self.stack.pop(0)
                    self.step_size += 1
                    debug_print(self.debug, "\n(pos={}, node_count={}) Current node.".format(pos, self.node_count))
                    timing = time.time()
                    solved = self.apply_and_check_goal(selection)
                    debug_print(self.debug, "(solved={}, timing={:.4f}s) Apply selection and check goal.".format(
                        solved, time.time() - timing))
                    if solved is None:  # not update, close search branch
                        continue
                    if solved:  # solved, return result
                        _, seqs = get_used_pid_and_theorem(self.problem)
                        return True, seqs
                    else:  # continue search
                        if len(pos) == self.max_depth:
                            continue
                        timing = time.time()
                        selections = self.get_theorem_selection()
                        self.add_selections(pos, selections)
                        debug_print(self.debug, "(timing={:.4f}s) Expand {} child node.".
                                    format(time.time() - timing, len(selections)))

        return False, None

    def get_theorem_selection(self):
        """
        Return theorem selections according to <self.last_step>.
        :return selections: <list> of ((t_name, t_branch, t_para), ((predicate, item, premise))).
        """
        selections = []

        timing = time.time()
        related_pres = []  # new added predicates
        related_syms = []  # new added/updated equations
        for step in range(self.last_step, self.problem.condition.step_count):  # get related conditions
            for _id in self.problem.condition.ids_of_step[step]:
                if self.problem.condition.items[_id][0] == "Equation":
                    for sym in self.problem.condition.items[_id][1].free_symbols:
                        if sym in related_syms:
                            continue
                        related_syms.append(sym)
                else:
                    if self.problem.condition.items[_id][0] not in self.p2t_map:
                        continue
                    item = self.problem.condition.items[_id][1]
                    for t_name, t_branch, p_vars in self.p2t_map[self.problem.condition.items[_id][0]]:
                        if len(p_vars) != len(item):
                            continue
                        letters = {}
                        for i in range(len(p_vars)):
                            letters[p_vars[i]] = item[i]
                        related_pre = (t_name, t_branch, letters)
                        if related_pre not in related_pres:
                            related_pres.append(related_pre)
        debug_print(self.debug, "(timing={:.4f}s) Get Related.".format(time.time() - timing))
        debug_print(self.debug, "Related predicates: {}.".format(related_pres))
        debug_print(self.debug, "Related syms: {}.".format(related_syms))

        timing = time.time()
        logic_selections = self.try_theorem_logic(related_pres)
        debug_print(self.debug, "(timing={:.4f}s) Get {} logic-related selections: {}.".format(
            time.time() - timing, len(logic_selections), logic_selections))
        timing = time.time()
        algebra_selections = self.try_theorem_algebra(related_syms)
        debug_print(self.debug, "(timing={:.4f}s) Get {} algebra-related selections: {}.".format(
            time.time() - timing, len(algebra_selections), algebra_selections))

        timing = time.time()
        added_selections = []
        for selection in logic_selections + algebra_selections:  # remove redundancy
            _, conclusions = selection
            s = []
            for conclusion in conclusions:
                predicate, item, _ = conclusion
                s.append((predicate, item))
            s = tuple(s)
            if s not in added_selections:
                added_selections.append(s)
                selections.append(selection)

        for i in range(len(selections))[::-1]:
            t_msg, conclusions = selections[i]
            t_name, t_branch, t_para = t_msg
            if "area" in t_name:
                if "ratio" in t_name:
                    para1 = t_para[0:int(len(t_para) / 2)]
                    para2 = t_para[int(len(t_para) / 2):]
                    if not (para1 in self.problem_a_paras and para2 in self.problem_a_paras):
                        selections.pop(i)
                else:
                    if t_para not in self.problem_a_paras:
                        selections.pop(i)
            elif "perimeter" in t_name:
                if "ratio" in t_name:
                    para1 = t_para[0:int(len(t_para) / 2)]
                    para2 = t_para[int(len(t_para) / 2):]
                    if not (para1 in self.problem_p_paras and para2 in self.problem_p_paras):
                        selections.pop(i)
                else:
                    if t_para not in self.problem_p_paras:
                        selections.pop(i)
        debug_print(self.debug, "(timing={:.4f}s) Get {}  selections: {}.".format(
            time.time() - timing, len(selections), selections))

        return selections

    def try_theorem_logic(self, related_pres):
        """
        Try a theorem and return can-added conclusions.
        :param related_pres: <list>, list of tuple('t_name', 't_branch', letters).
        :return selections: <list> of ((t_name, t_branch, t_para, t_timing), ((predicate, item, premise))).
        """

        selections = []
        for t_name, t_branch, t_letters in related_pres:
            gpl = self.parsed_theorem_GDL[t_name]["body"][t_branch]
            results = GPLExecutor.run(gpl, self.problem, t_letters)  # get gpl reasoned result
            for letters, premise, conclusion in results:
                t_para = tuple([letters[i] for i in self.parsed_theorem_GDL[t_name]["vars"]])

                premise = tuple(premise)
                conclusions = []
                for predicate, item in conclusion:  # add conclusion
                    if self.problem.check(predicate, item, premise, t_name):
                        if predicate != "Equation":
                            item = tuple(item)
                        conclusions.append((predicate, item, premise))

                if len(conclusions) > 0:
                    selections.append(((t_name, t_branch, t_para), tuple(conclusions)))

        return selections

    def try_theorem_algebra(self, related_syms):
        """
        Try a theorem and return can-added conclusions.
        :param related_syms: <list>, related syms.
        :return selections: <list> of ((t_name, t_branch, t_para, t_timing), ((predicate, item, premise))).
        """
        paras_of_attrs = {}  # <dict>, {attr: [para]}
        for sym in related_syms:
            attr, paras = self.problem.condition.attr_of_sym[sym]
            if attr not in self.p2t_map:
                continue

            if attr not in paras_of_attrs:
                paras_of_attrs[attr] = []

            for para in paras:
                if para in paras_of_attrs[attr]:
                    continue
                paras_of_attrs[attr].append(para)

        selections = []
        for related_attr in paras_of_attrs:
            related_paras = set(paras_of_attrs[related_attr])
            for t_name, t_branch, p_vars in self.p2t_map[related_attr]:
                gpl = self.parsed_theorem_GDL[t_name]["body"][t_branch]  # run gdl
                for related_para in related_paras:
                    letters = {}
                    for i in range(len(p_vars)):
                        letters[p_vars[i]] = related_para[i]
                    results = GPLExecutor.run(gpl, self.problem, letters)  # get gpl reasoned result
                    for letters, premise, conclusion in results:
                        theorem_para = tuple([letters[i] for i in self.parsed_theorem_GDL[t_name]["vars"]])
                        premise = tuple(premise)
                        conclusions = []
                        for predicate, item in conclusion:  # add conclusion
                            if self.problem.check(predicate, item, premise, t_name):
                                if predicate != "Equation":
                                    item = tuple(item)
                                conclusions.append((predicate, item, premise))

                        if len(conclusions) > 0:
                            selections.append(((t_name, t_branch, theorem_para), tuple(conclusions)))
        return selections

    def apply_and_check_goal(self, selection):
        """
        Apply selection and check goal.
        :param selection: ((t_name, t_branch, t_para), ((predicate, item, premise))).
        :return solved: <bool> or None. Set None when not update
        """
        self.last_step = self.problem.condition.step_count
        update = False
        t_msg, conclusions = selection

        for predicate, item, premise in conclusions:
            update = self.problem.add(predicate, item, premise, t_msg, skip_check=True) or update

        if not update:  # close current branch if applied theorem no new condition
            return None

        EqKiller.solve_equations(self.problem)  # solve eq & check_goal
        self.problem.check_goal()
        self.problem.step(t_msg, 0)

        return self.problem.goal.solved

    def add_selections(self, father_pos, selections):
        """
        Add selections to self.stack.
        :param father_pos: position of father branch.
        :param selections: <list> of ((t_name, t_branch, t_para), ((predicate, item, premise))).
        """
        pos = list(father_pos)
        depth = len(father_pos) + 1
        if depth not in self.node_count:
            self.node_count[depth] = 1

        for selection in selections:
            self.stack.append((tuple(pos + [self.node_count[depth]]), selection))
            self.node_count[depth] += 1



import json
import os


def load_json(file_path_and_name):
    with open(file_path_and_name, "r", encoding="utf-8") as f:
        return json.load(f)


def save_json(data, file_path_and_name):
    with open(file_path_and_name, "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=2)


def safe_save_json(data, file_path_and_name):
    """Avoiding log file corruption."""
    path = os.path.dirname(file_path_and_name)
    filename = os.path.basename(file_path_and_name)
    file_path_and_name_bk = os.path.join(path, filename + ".bk")

    with open(file_path_and_name_bk, "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=2)

    if os.path.exists(file_path_and_name):
        os.remove(file_path_and_name)

    os.rename(file_path_and_name_bk, file_path_and_name)


def debug_print(debug, msg):
    """Print 'msg' when debug is True."""
    if debug:
        print(msg)


def rough_equal(a, b):
    """Accuracy is controlled at 0.001. Preventing floating point calculation issues"""
    return abs(a - b) < 0.001


def get_user_input(notes, choice=None):
    if choice is None:
        choice = ["y", "n"]
    notes = "{} ({}):".format(notes, "/".join(choice))
    user_input = input(notes)
    while user_input not in choice:
        user_input = input(notes)
    return user_input



# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""
'Problem' preserves all details of the problem-solving process, ensures the correctness and
consistency of the problem input conditions, and implements automatic diagram construction,
condition auto-expansion, and validity checks.
"""

__all__ = [
    "Problem"
]

from formalgeo.problem.problem import Problem



from sympy import Float
from formalgeo.parse import inverse_parse_one, inverse_parse_logic_to_cdl, inverse_parse_one_theorem
from formalgeo.tools import save_json
from graphviz import Digraph, Graph
import os


def simple_show(problem, timing):
    """Show simple information about problem-solving."""
    pid = problem.parsed_problem_CDL["id"]
    correct_answer = str(problem.goal.answer).replace(" ", "")
    solved = problem.goal.solved
    solved_answer = problem.goal.solved_answer

    solved = "\033[32msolved\033[0m\t" if solved else "\033[31munsolved\033[0m\t"
    timing = "{:.6f}".format(timing) if timing < 2 else "\033[31m{:.6f}\033[0m".format(timing)
    print("{}\t{}\t{}\t{}\t{}".format(pid, str(correct_answer), solved, solved_answer, timing))


def show_solution(problem):
    """-----------Condition Declaration Statements-----------"""
    print("\033[36mproblem_index:\033[0m", end=" ")
    print(problem.parsed_problem_CDL["id"])
    print("\033[36mconstruction_cdl:\033[0m")
    for construction_fl in problem.parsed_problem_CDL["cdl"]["construction_cdl"]:
        print(construction_fl)
    print("\033[36mtext_cdl:\033[0m")
    for text_fl in problem.parsed_problem_CDL["cdl"]["text_cdl"]:
        print(text_fl)
    print("\033[36mimage_cdl:\033[0m")
    for image_fl in problem.parsed_problem_CDL["cdl"]["image_cdl"]:
        print(image_fl)
    print("\033[36mgoal_cdl:\033[0m")
    print(problem.parsed_problem_CDL["cdl"]["goal_cdl"])
    print()

    """-----------Process of Problem Solving-----------"""
    print("\033[36mreasoning:\033[0m")
    anti_parsed_cdl = inverse_parse_logic_to_cdl(problem)
    for step in anti_parsed_cdl:
        for cdl in anti_parsed_cdl[step]:
            print("{0:^3}{1:<20}".format(step, cdl))
    print()

    used_pid, used_theorem = get_used_pid_and_theorem(problem)

    """-----------Logic Form-----------"""
    print("\033[33mRelations:\033[0m")
    basic_predicates = list(problem.parsed_predicate_GDL["Preset"]["Construction"])
    basic_predicates += list(problem.parsed_predicate_GDL["Preset"]["BasicEntity"])
    predicates = list(problem.parsed_predicate_GDL["Entity"])
    predicates += list(problem.parsed_predicate_GDL["Relation"])
    for predicate in basic_predicates + predicates:
        items = problem.condition.get_items_by_predicate(predicate)
        if len(items) == 0:
            continue
        print(predicate + ":")

        if predicate in basic_predicates and len(items) > 20:
            items = items[0:10] + ["..."] + items[len(items) - 10:]

        for item in items:
            if isinstance(item, str):
                print(item)
                continue
            _id = problem.condition.get_id_by_predicate_and_item(predicate, item)
            if len(problem.condition.items[_id][2]) <= 3:
                premises = "(" + ",".join([str(i) for i in problem.condition.items[_id][2]]) + ")"
            else:
                premises = "(" + ",".join([str(i) for i in problem.condition.items[_id][2][0:3]]) + ",...)"
            theorem = inverse_parse_one_theorem(problem.condition.items[_id][3], problem.parsed_theorem_GDL)
            item = inverse_parse_one(predicate, item, problem, True)
            if len(item) > 35:
                item = item[0:35] + "..."
            if _id not in used_pid:
                print("{0:^6}{1:^50}{2:^25}{3:^6}".format(_id, item, premises, theorem))
            else:
                print("\033[35m{0:^6}{1:^50}{2:^25}{3:^6}\033[0m".format(_id, item, premises, theorem))
    print()

    print("\033[33mSymbols and Value:\033[0m")
    for attr in problem.condition.sym_of_attr:
        sym = problem.condition.sym_of_attr[attr]
        if isinstance(problem.condition.value_of_sym[sym], Float):
            print("{0:^70}{1:^15}{2:^20.3f}".format(
                "{}({})".format(attr[0], "".join(attr[1])), str(sym), problem.condition.value_of_sym[sym]))
        else:
            print("{0:^70}{1:^15}{2:^20}".format(
                "{}({})".format(attr[0], "".join(attr[1])), str(sym), str(problem.condition.value_of_sym[sym])))

    print("\033[33mEquations:\033[0m")
    items = problem.condition.get_items_by_predicate("Equation")
    for item in items:
        _id = problem.condition.get_id_by_predicate_and_item("Equation", item)
        if len(problem.condition.items[_id][2]) <= 3:
            premises = "(" + ",".join([str(i) for i in problem.condition.items[_id][2]]) + ")"
        else:
            premises = "(" + ",".join([str(i) for i in problem.condition.items[_id][2][0:3]]) + ",...)"
        theorem = inverse_parse_one_theorem(problem.condition.items[_id][3], problem.parsed_theorem_GDL)
        item = str(item).replace(" ", "")
        if len(item) > 40:
            item = item[0:40] + "..."
        if _id not in used_pid:
            print("{0:^6}{1:^60}{2:^25}{3:>6}".format(_id, item, premises, theorem))
        else:
            print("\033[35m{0:^6}{1:^60}{2:^25}{3:>6}\033[0m".format(_id, item, premises, theorem))
    print()

    # goal
    print("\033[34mSolving Goal:\033[0m")
    print("type: {}".format(problem.goal.type))
    print("goal: {}".format(str(problem.goal.item).replace(" ", "")))
    print("answer: {}".format(str(problem.goal.answer).replace(" ", "")))
    if problem.goal.solved:
        print("solved: \033[32mTrue\033[0m")
    else:
        print("solved: \033[31mFalse\033[0m")
    if problem.goal.solved_answer is not None:
        print("solved_answer: {}".format(str(problem.goal.solved_answer)))
        if not isinstance(problem.goal.solved_answer, tuple):
            try:
                print("solved_answer(float): {}".format(float(problem.goal.solved_answer)))
            except TypeError:
                print("solved_answer(float): <Cannot convert to float>")
    if problem.goal.premise is not None:
        print("premise: {}".format(str(problem.goal.premise)))
        print("theorem: {}".format(inverse_parse_one_theorem(problem.goal.theorem, problem.parsed_theorem_GDL)))
    print()

    print("\033[34mTiming:\033[0m")
    time_sum = 0
    for step in problem.timing:
        if problem.timing[step][0] in used_theorem:
            theorem = inverse_parse_one_theorem(problem.timing[step][0], problem.parsed_theorem_GDL)
            print("\033[35m{:^2} {} {:.6f}s\033[0m".format(step, theorem, problem.timing[step][1]))
        else:
            print("{:^2} {} {:.6f}s".format(step, problem.timing[step][0], problem.timing[step][1]))
        time_sum += problem.timing[step][1]
    print("total: {:.6f}s".format(time_sum))


def get_solution_step(problem):
    """Get conditions grouped by step in dict."""
    step_msg = {
        "cdl": inverse_parse_logic_to_cdl(problem),
        "theorems_applied": {}
    }
    for step in problem.timing:
        if isinstance(problem.timing[step][0], tuple):
            step_msg["theorems_applied"][str(step)] = inverse_parse_one_theorem(
                problem.timing[step][0], problem.parsed_theorem_GDL)

    return step_msg


def get_solution_hypertree(problem):
    """Generate solution hyper tree and theorem DAG."""
    group = {}  # (premise, theorem): [_id], used for building hyper graph.
    cdl = {}  # _id: anti_parsed_cdl, user for getting cdl by id.

    for _id in range(problem.condition.id_count):  # summary information
        predicate, item, premise, theorem, _ = problem.condition.items[_id]
        theorem = inverse_parse_one_theorem(theorem, problem.parsed_theorem_GDL)
        cdl[_id] = inverse_parse_one(predicate, item, problem)
        if theorem == "prerequisite":  # prerequisite not show in graph
            continue
        if (premise, theorem) not in group:
            group[(premise, theorem)] = [_id]
        else:
            group[(premise, theorem)].append(_id)

    if problem.goal.solved and problem.goal.type == "algebra":  # if target solved, add target
        eq = problem.goal.item - problem.goal.answer
        if eq not in problem.condition.get_items_by_predicate("Equation"):  # target not in condition set
            target_equation = inverse_parse_one("Equation", eq, problem)
            _id = len(cdl)
            cdl[_id] = target_equation
            group[(problem.goal.premise, problem.goal.theorem)] = [_id]

    solution_hypertree = {}
    for premise, theorem in group:  # generate solution tree
        start_nodes = [cdl[_id] for _id in premise]

        conclusion = group[(premise, theorem)]
        end_nodes = [cdl[_id] for _id in conclusion]

        solution_hypertree[len(solution_hypertree)] = {
            "conditions": start_nodes,
            "theorem": theorem,
            "conclusion": end_nodes
        }

    return solution_hypertree


def get_theorem_dag(problem):
    """Generate theorem DAG."""
    group = {}  # (premise, theorem): [_id], used for building theorem DAG.
    for _id in range(problem.condition.id_count):  # summary information
        predicate, item, premise, theorem, _ = problem.condition.items[_id]
        if theorem[0] == "prerequisite":  # prerequisite not show in graph
            continue
        if (premise, theorem) not in group:
            group[(premise, theorem)] = [_id]
        else:
            group[(premise, theorem)].append(_id)
    if problem.goal.solved and problem.goal.type == "algebra":  # if target solved, add target
        eq = problem.goal.item - problem.goal.answer
        if eq not in problem.condition.get_items_by_predicate("Equation"):  # target not in condition set
            group[(problem.goal.premise, problem.goal.theorem)] = [problem.condition.id_count]

    adjust_group = {}
    for premise, theorem in group:
        adjust_theorem = (theorem[0], theorem[1], theorem[2], len(adjust_group))
        adjust_group[(premise, adjust_theorem)] = group[(premise, theorem)]

    rough_dag = {}
    for premise, theorem in adjust_group:  # generate theorem dag
        conclusion = adjust_group[(premise, theorem)]
        if theorem not in rough_dag:
            rough_dag[theorem] = []

        for _id in conclusion:
            for tail_premise, tail_theorem in adjust_group:
                if _id in tail_premise and tail_theorem not in rough_dag[theorem]:
                    rough_dag[theorem].append(tail_theorem)

    update = True
    while update:  # remove tail which contains 'extended' and 'solve_eq'
        update = False
        for head in rough_dag:
            for tail in list(rough_dag[head]):
                if tail[0] not in ["extended", "solve_eq"]:
                    continue
                rough_dag[head].pop(rough_dag[head].index(tail))
                if tail not in rough_dag:
                    continue
                for new_tail in rough_dag[tail]:
                    rough_dag[head].append(new_tail)
                    update = True

    cleaned_dag = {}
    for head in rough_dag:  # remove head which contains 'extended' and 'solve_eq'
        if head[0] in ["extended", "solve_eq"]:
            continue
        cleaned_dag[head] = []
        for tail in rough_dag[head]:
            cleaned_dag[head].append(tail)

    start_nodes = []
    for head in cleaned_dag:  # get START nodes
        is_start_node = True
        for check_head in cleaned_dag:
            if head in cleaned_dag[check_head]:
                is_start_node = False
                break
        if is_start_node:
            start_nodes.append(inverse_parse_one_theorem(
                (head[0], head[1], head[2]), problem.parsed_theorem_GDL))

    dag = {"START": start_nodes}
    for head in cleaned_dag:
        if len(cleaned_dag[head]) == 0:
            continue
        inverse_adjust_head = inverse_parse_one_theorem(
            (head[0], head[1], head[2]), problem.parsed_theorem_GDL)
        dag[inverse_adjust_head] = [
            inverse_parse_one_theorem((tail[0], tail[1], tail[2]), problem.parsed_theorem_GDL)
            for tail in cleaned_dag[head]
        ]

    return dag


def get_used_pid_and_theorem(problem):
    """Return used condition id and theorem for solving problem."""
    if not problem.goal.solved:
        return [], []

    used_pid = list(set(problem.goal.premise))
    used_theorem = []
    while True:
        len_used_pid = len(used_pid)
        for _id in used_pid:
            if _id >= 0:
                used_pid += list(problem.condition.items[_id][2])
                used_theorem.append(problem.condition.items[_id][3])
        used_pid = list(set(used_pid))  # 快速去重
        if len_used_pid == len(used_pid):
            break

    selected_theorem = []
    for step in problem.timing:  # ensure ordered theorem seqs list
        if problem.timing[step][0] in used_theorem and \
                problem.timing[step][0] not in selected_theorem and \
                problem.timing[step][0][0] not in ["solve_eq", "extended", "prerequisite"]:
            selected_theorem.append(problem.timing[step][0])
    if problem.goal.theorem[0] not in ["solve_eq", "extended", "prerequisite"] and \
            problem.goal.theorem not in selected_theorem:
        selected_theorem.append(problem.goal.theorem)

    return used_pid, selected_theorem


def draw_solution_tree_and_theorem_dag(problem, path):
    """Generate and save solution hyper tree and theorem DAG."""
    st_dot = Digraph(name=str(problem.parsed_problem_CDL["id"]))  # Tree
    nodes = []  # list of node(cdl or theorem).
    t_nodes = []  # theorem nodes, used for DAG generating.
    edges = {}  # node(cdl or theorem): [node(cdl or theorem)], used for DAG generating.
    group = {}  # (premise, theorem): [_id], used for building hyper graph.
    cdl = {}  # _id: anti_parsed_cdl, user for getting cdl by id.

    for _id in range(problem.condition.id_count):  # summary information
        predicate, item, premise, theorem, _ = problem.condition.items[_id]
        theorem = inverse_parse_one_theorem(theorem, problem.parsed_theorem_GDL)
        cdl[_id] = inverse_parse_one(predicate, item, problem)
        if theorem == "prerequisite":  # prerequisite not show in graph
            continue
        if (premise, theorem) not in group:
            group[(premise, theorem)] = [_id]
        else:
            group[(premise, theorem)].append(_id)

    used_id, _ = get_used_pid_and_theorem(problem)  # just keep useful solution msg
    if problem.goal.solved:  # if target solved, add target
        if problem.goal.type == "algebra":
            eq = problem.goal.item - problem.goal.answer
            if eq not in problem.condition.get_items_by_predicate("Equation"):  # target not in condition set
                target_equation = inverse_parse_one("Equation", eq, problem)
                _id = len(cdl)
                cdl[_id] = target_equation
                group[(problem.goal.premise, problem.goal.theorem)] = [_id]
                used_id.append(_id)
            else:
                used_id.append(problem.condition.get_id_by_predicate_and_item("Equation", eq))
        else:
            used_id.append(problem.condition.get_id_by_predicate_and_item(problem.goal.item, problem.goal.answer))
        used_id += list(problem.goal.premise)

    remove_list = []
    for key in group:
        premise = key[0]
        not_used_in_premise = True
        l = 0
        while not_used_in_premise and l < len(premise):
            if premise[l] in used_id:
                not_used_in_premise = False
                break
            l += 1

        conclusion = group[key]
        not_used_in_conclusion = True
        l = 0
        while not_used_in_conclusion and l < len(conclusion):
            if conclusion[l] in used_id:
                not_used_in_conclusion = False
                break
            l += 1

        if not_used_in_premise or not_used_in_conclusion:
            remove_list.append(key)

    for key in remove_list:
        group.pop(key)

    count = 0
    solution_tree = {}
    for key in group:  # generate solution tree
        premise, theorem = key
        conclusion = group[key]

        theorem_node = theorem + "_{}".format(count)  # theorem name in hyper
        t_nodes.append(theorem_node)
        _add_node(st_dot, nodes, theorem_node)

        start_nodes = []
        for _id in premise:
            if _id not in used_id:
                continue
            _add_node(st_dot, nodes, cdl[_id])  # add node to graph
            start_nodes.append(cdl[_id])  # add to json output
            _add_edge(st_dot, nodes, cdl[_id], theorem_node, edges)  # add edge to graph

        end_nodes = []
        for _id in conclusion:
            if _id not in used_id:
                continue
            _add_node(st_dot, nodes, cdl[_id])  # add node to graph
            end_nodes.append(cdl[_id])  # add to json output
            _add_edge(st_dot, nodes, theorem_node, cdl[_id], edges)  # add edge to graph

        solution_tree[count] = {
            "conditions": start_nodes,
            "theorem": theorem,
            "conclusion": end_nodes
        }
        count += 1

    save_json(solution_tree, path + "{}_hyper.json".format(problem.parsed_problem_CDL["id"]))  # save solution tree
    st_dot.render(directory=path, view=False, format="png")  # save hyper graph
    os.remove(path + "{}.gv".format(problem.parsed_problem_CDL["id"]))
    if "{}_hyper.png".format(problem.parsed_problem_CDL["id"]) in os.listdir(path):
        os.remove(path + "{}_hyper.png".format(problem.parsed_problem_CDL["id"]))
    os.rename(path + "{}.gv.png".format(problem.parsed_problem_CDL["id"]),
              path + "{}_hyper.png".format(problem.parsed_problem_CDL["id"]))

    dag_dot = Digraph(name=str(problem.parsed_problem_CDL["id"]))  # generate theorem DAG
    nodes = []  # list of theorem.
    dag = {}

    for s_node in edges:  # select theorem nodes
        if s_node not in t_nodes:
            continue
        dag[s_node] = []
        for m_node in edges[s_node]:  # middle condition
            if m_node not in edges:
                continue
            for e_node in edges[m_node]:
                dag[s_node].append(e_node)

    update = True
    while update:  # remove extended and solve_eq nodes
        update = False
        for head in dag:
            for tail in dag[head]:
                if not (tail.startswith("extended") or tail.startswith("solve_eq")):
                    continue
                dag[head].pop(dag[head].index(tail))
                if tail not in dag:
                    continue
                for new_tail in dag[tail]:
                    dag[head].append(new_tail)
                    update = True

    cleaned = {}
    for key in dag:
        if key.startswith("extended") or key.startswith("solve_eq"):
            continue
        new_key = key.split(")")[0] + ")"
        cleaned[new_key] = []
        for tail in dag[key]:
            cleaned[new_key].append(tail.split(")")[0] + ")")
    dag = cleaned

    root_nodes = []
    child_nodes = []
    real_root_nodes = []
    real_child_nodes = []
    for head in dag:  # build DAG graph
        _add_node(dag_dot, nodes, head)
        root_nodes.append(head)
        if len(dag[head]) == 0:
            real_child_nodes.append(head)
            continue
        for tail in set(dag[head]):
            _add_node(dag_dot, nodes, tail)
            _add_edge(dag_dot, nodes, head, tail)
            child_nodes.append(tail)

    _add_node(dag_dot, nodes, "START")  # add START node
    for root in root_nodes:
        if root in child_nodes:
            continue
        real_root_nodes.append(root)
        _add_node(dag_dot, nodes, root)
        _add_edge(dag_dot, nodes, "START", root)
    dag["START"] = real_root_nodes

    save_json(dag, path + "{}_dag.json".format(problem.parsed_problem_CDL["id"]))  # save solution tree
    dag_dot.render(directory=path, view=False, format="png")  # save hyper graph
    os.remove(path + "{}.gv".format(problem.parsed_problem_CDL["id"]))
    if "{}_dag.png".format(problem.parsed_problem_CDL["id"]) in os.listdir(path):
        os.remove(path + "{}_dag.png".format(problem.parsed_problem_CDL["id"]))
    os.rename(path + "{}.gv.png".format(problem.parsed_problem_CDL["id"]),
              path + "{}_dag.png".format(problem.parsed_problem_CDL["id"]))


def _add_node(dot, nodes, node):
    if node in nodes:  # node was already added
        return

    added_node_id = len(nodes)
    nodes.append(node)
    if node[0].isupper():
        dot.node(str(added_node_id), node, shape='box')  # condition node
    else:
        dot.node(str(added_node_id), node)  # theorem node


def _add_edge(dot, nodes, start_node, end_node, edges=None):
    dot.edge(str(nodes.index(start_node)), str(nodes.index(end_node)))
    if edges is not None:
        if start_node not in edges:
            edges[start_node] = [end_node]
        else:
            edges[start_node].append(end_node)
